@@ -53,6 +53,14 @@ static void mb_set_cdrom (SjMetadata *metadata, const char* device);
 static void mb_set_proxy (SjMetadata *metadata, const char* proxy);
 static void mb_set_proxy_port (SjMetadata *metadata, const int port);
 static void mb_list_albums (SjMetadata *metadata, GError **error);
+static char *mb_get_submit_url (SjMetadata *metadata);
+
+#define GCONF_PROXY_USE_PROXY "/system/http_proxy/use_http_proxy"
+#define GCONF_PROXY_HOST "/system/http_proxy/host"
+#define GCONF_PROXY_PORT "/system/http_proxy/port"
+#define GCONF_PROXY_USE_AUTHENTICATION "/system/http_proxy/use_authentication"
+#define GCONF_PROXY_USERNAME "/system/http_proxy/authentication_user"
+#define GCONF_PROXY_PASSWORD "/system/http_proxy/authentication_password"
 
 /**
  * GObject methods
@@ -100,9 +108,27 @@ sj_metadata_musicbrainz_instance_init (GTypeInstance *instance, gpointer g_class
     mb_SetServer (self->priv->mb, server_name, 80);
     g_free (server_name);
   }
+  
+  /* Set the HTTP proxy */
+  if (gconf_client_get_bool (gconf_client, GCONF_PROXY_USE_PROXY, NULL)) {
+    mb_SetProxy (self->priv->mb,
+                 gconf_client_get_string (gconf_client, GCONF_PROXY_HOST, NULL),
+                 gconf_client_get_int (gconf_client, GCONF_PROXY_PORT, NULL));
+    if (gconf_client_get_bool (gconf_client, GCONF_PROXY_USE_AUTHENTICATION, NULL)) {
+#if HAVE_MB_SETPROXYCREDS
+      mb_SetProxyCreds (self->priv->mb,
+                        gconf_client_get_string (gconf_client, GCONF_PROXY_USERNAME, NULL),
+                        gconf_client_get_string (gconf_client, GCONF_PROXY_USERNAME, NULL));
+#else
+      g_warning ("mb_SetProxyCreds() not found, no proxy authorisation possible.");
+#endif
+    }
+  }
+
   g_object_unref (gconf_client);
 
-  /* TODO: optimal setting? mb_SetDepth (self->priv->mb, 1); */
+  //mb_SetDepth (self->priv->mb, 1);
+
   if (g_getenv("MUSICBRAINZ_DEBUG")) {
     mb_SetDebug (self->priv->mb, TRUE);
   }
@@ -117,6 +143,7 @@ metadata_interface_init (gpointer g_iface, gpointer iface_data)
   klass->set_proxy = mb_set_proxy;
   klass->set_proxy_port = mb_set_proxy_port;
   klass->list_albums = mb_list_albums;
+  klass->get_submit_url = mb_get_submit_url;
 }
 
 static void
@@ -282,15 +309,159 @@ mb_set_proxy_port (SjMetadata *metadata, const int port)
   mb_SetProxy (priv->mb, priv->http_proxy, priv->http_proxy_port);
 }
 
+/* Data imported from FreeDB is horrendeous for compilations,
+ * Try to split the 'Various' artist */
+static void
+artist_and_title_from_title (TrackDetails *track, gpointer data)
+{
+  char *slash, **split;
+
+  if (g_ascii_strncasecmp (MBI_VARIOUS_ARTIST_ID, track->album->artist_id, 64) != 0 && track->album->artist_id[0] != '\0' && track->artist_id[0] != '\0') {
+    track->title = g_strdup (data);
+    return;
+  }
+
+  slash = strstr (data, " / ");
+  if (slash == NULL) {
+    track->title = g_strdup (data);
+    return;
+  }
+  split = g_strsplit (data, " / ", 2);
+  track->artist = g_strdup (split[0]);
+  track->title = g_strdup (split[1]);
+  g_strfreev (split);
+}
+
+#if WITH_CACHE
+/**
+ * Write the RDF in the MusicBrainz object to the file specified.
+ */
+static void
+cache_rdf (musicbrainz_t mb, const char *filename)
+{
+  GError *error = NULL;
+  int len;
+  char *path, *rdf;
+
+  g_assert (mb != NULL);
+  g_assert (filename != NULL);
+
+  /* Create the folder for the file */
+  path = g_path_get_dirname (filename);
+  g_mkdir_with_parents (path, 0755); /* Handle errors in set_contents() */
+  g_free (path);
+  
+  /* How much data is there to save? */
+  len = mb_GetResultRDFLen (mb);
+  rdf = g_malloc0 (len);
+
+  /* Get the RDF and save it */
+  mb_GetResultRDF (mb, rdf, len);
+  if (!g_file_set_contents (filename, rdf, len, &error)) {
+    g_warning ("Cannot write cache file %s: %s", filename, error->message);
+    g_error_free (error);
+  }
+
+  g_free (rdf);
+}
+
+/**
+ * Load into the MusicBrainz object the RDF from the specified cache file if it
+ * exists and is valid then return TRUE, otherwise return FALSE.
+ */
+static gboolean
+get_cached_rdf (musicbrainz_t mb, const char *cachepath)
+{
+  gboolean ret = FALSE;
+  GError *error = NULL;
+  char *rdf = NULL;
+
+  g_assert (mb != NULL);
+  g_assert (cachepath != NULL);
+
+  if (!g_file_test (cachepath, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR))
+    goto done;
+  
+  /* Cache file exists, open it */
+  if (!g_file_get_contents (cachepath, &rdf, NULL, &error)) {
+    g_warning ("Cannot open cache file %s: %s", cachepath, error->message);
+    g_error_free (error);
+    goto done;
+  }
+  
+  /* Set the RDF */
+  if (mb_SetResultRDF (mb, rdf))
+    ret = TRUE;
+  
+  done:
+    g_free (rdf);
+    return ret;
+}
+#else
+static gboolean
+get_cached_rdf (musicbrainz_t mb, const char *cachepath)
+{
+  return FALSE;
+}
+static void
+cache_rdf (musicbrainz_t mb, const char *filename) {
+}
+#endif
+
+/**
+ * Fill the MusicBrainz object with RDF.  Basically get the CD Index and check
+ * the local cache, if that fails then lookup the data online.
+ */
+static void
+get_rdf (SjMetadata *metadata)
+{
+  SjMetadataMusicbrainzPrivate *priv;
+  char data[256];
+  char *cdindex, *cachepath;
+
+  g_assert (metadata != NULL);
+
+  priv = SJ_METADATA_MUSICBRAINZ (metadata)->priv;
+
+  /* Get the Table of Contents */
+  if (!mb_Query (priv->mb, MBQ_GetCDTOC)) {
+    mb_GetQueryError (priv->mb, data, sizeof (data));
+    g_print (_("This CD could not be queried: %s\n"), data);
+    return;
+  }
+
+  /* Extract the CD Index */
+  if (!mb_GetResultData(priv->mb, MBE_TOCGetCDIndexId, data, sizeof (data))) {
+    mb_GetQueryError (priv->mb, data, sizeof (data));
+    g_print (_("This CD could not be queried: %s\n"), data);
+    return;
+  }
+  cdindex = g_strdup (data);
+
+  cachepath = g_build_filename (g_get_home_dir (), ".gnome2", "sound-juicer", "cache", cdindex, NULL);
+  
+  if (!get_cached_rdf (priv->mb, cachepath)) {
+    if (!mb_QueryWithArgs (priv->mb, MBQ_GetCDInfoFromCDIndexId, (char*[]) {cdindex, NULL})) {
+      mb_GetQueryError (priv->mb, data, sizeof (data));
+      g_print (_("This CD could not be queried: %s\n"), data);
+      goto done;
+    }
+    cache_rdf (priv->mb, cachepath);
+  }
+  
+ done:
+  g_free (cachepath);
+  g_free (cdindex);
+}
+
 static gpointer
 lookup_cd (SjMetadata *metadata)
 {
   /** The size of the buffer used in MusicBrainz lookups */
-#define MB_BUFFER_SIZE 256
   SjMetadataMusicbrainzPrivate *priv;
   GList *albums = NULL;
   GList *al, *tl;
-  char data[MB_BUFFER_SIZE];
+  char data[256];
   int num_albums, i, j;
   NautilusBurnMediaType type;
 
@@ -321,17 +492,10 @@ lookup_cd (SjMetadata *metadata)
     return NULL;
   }
 
-  if (!mb_Query (priv->mb, MBQ_GetCDInfo)) {
-    mb_GetQueryError (priv->mb, data, MB_BUFFER_SIZE);
-    g_print (_("This CD could not be queried: %s\n"), data);
-    priv->albums = get_offline_track_listing (metadata, &(priv->error));
-    g_idle_add ((GSourceFunc)fire_signal_idle, metadata);
-    return priv->albums;
-  }
+  get_rdf (metadata);
 
   num_albums = mb_GetResultInt(priv->mb, MBE_GetNumAlbums);
   if (num_albums < 1) {
-    g_print (_("This CD was not found.\n"));
     priv->albums = get_offline_track_listing (metadata, &(priv->error));
     g_idle_add ((GSourceFunc)fire_signal_idle, metadata);
     return priv->albums;
@@ -344,26 +508,32 @@ lookup_cd (SjMetadata *metadata)
     mb_Select1(priv->mb, MBS_SelectAlbum, i);
     album = g_new0 (AlbumDetails, 1);
 
-    if (mb_GetResultData(priv->mb, MBE_AlbumGetAlbumId, data, MB_BUFFER_SIZE)) {
-      mb_GetIDFromURL (priv->mb, data, data, MB_BUFFER_SIZE);
+    if (mb_GetResultData(priv->mb, MBE_AlbumGetAlbumId, data, sizeof (data))) {
+      mb_GetIDFromURL (priv->mb, data, data, sizeof (data));
       album->album_id = g_strdup (data);
     }
 
-    if (mb_GetResultData (priv->mb, MBE_AlbumGetAlbumArtistId, data, MB_BUFFER_SIZE)) {
-      mb_GetIDFromURL (priv->mb, data, data, MB_BUFFER_SIZE);
+    if (mb_GetResultData (priv->mb, MBE_AlbumGetAlbumArtistId, data, sizeof (data))) {
+      mb_GetIDFromURL (priv->mb, data, data, sizeof (data));
       album->artist_id = g_strdup (data);
       if (g_ascii_strncasecmp (MBI_VARIOUS_ARTIST_ID, data, 64) == 0) {
         album->artist = g_strdup (_("Various"));
+        album->artist_sortname = g_strdup (album->artist);
       } else {
-        if (data && mb_GetResultData1(priv->mb, MBE_AlbumGetArtistName, data, MB_BUFFER_SIZE, 1)) {
+        if (data && mb_GetResultData1(priv->mb, MBE_AlbumGetArtistName, data, sizeof (data), 1)) {
           album->artist = g_strdup (data);
         } else {
           album->artist = g_strdup (_("Unknown Artist"));
         }
+        if (data && mb_GetResultData1(priv->mb, MBE_AlbumGetArtistSortName, data, sizeof (data), 1)) {
+          album->artist_sortname = g_strdup (data);
+        } else {
+          album->artist_sortname = g_strdup (album->artist);
+        }
       }
     }
 
-    if (mb_GetResultData(priv->mb, MBE_AlbumGetAlbumName, data, MB_BUFFER_SIZE)) {
+    if (mb_GetResultData(priv->mb, MBE_AlbumGetAlbumName, data, sizeof (data))) {
       album->title = g_strdup (data);
     } else {
       album->title = g_strdup (_("Unknown Title"));
@@ -374,7 +544,7 @@ lookup_cd (SjMetadata *metadata)
       num_releases = mb_GetResultInt (priv->mb, MBE_AlbumGetNumReleaseDates);
       if (num_releases > 0) {
         mb_Select1(priv->mb, MBS_SelectReleaseDate, 1);
-        if (mb_GetResultData(priv->mb, MBE_ReleaseGetDate, data, MB_BUFFER_SIZE)) {
+        if (mb_GetResultData(priv->mb, MBE_ReleaseGetDate, data, sizeof (data))) {
           int matched, year=1, month=1, day=1;
           matched = sscanf(data, "%u-%u-%u", &year, &month, &day);
           if (matched > 1) {
@@ -388,9 +558,10 @@ lookup_cd (SjMetadata *metadata)
     num_tracks = mb_GetResultInt(priv->mb, MBE_AlbumGetNumTracks);
     if (num_tracks < 1) {
       g_free (album->artist);
+      g_free (album->artist_sortname);
       g_free (album->title);
       g_free (album);
-      g_print (_("Incomplete metadata for this CD.\n"));
+      g_warning (_("Incomplete metadata for this CD"));
       priv->albums = get_offline_track_listing (metadata, &(priv->error));
       g_idle_add ((GSourceFunc)fire_signal_idle, metadata);
       return priv->albums;
@@ -404,25 +575,33 @@ lookup_cd (SjMetadata *metadata)
 
       track->number = j; /* replace with number lookup? */
 
-      if (mb_GetResultData1(priv->mb, MBE_AlbumGetTrackId, data, MB_BUFFER_SIZE, j)) {
-        mb_GetIDFromURL (priv->mb, data, data, MB_BUFFER_SIZE);
+      if (mb_GetResultData1(priv->mb, MBE_AlbumGetTrackId, data, sizeof (data), j)) {
+        mb_GetIDFromURL (priv->mb, data, data, sizeof (data));
         track->track_id = g_strdup (data);
       }
 
-      if (mb_GetResultData1(priv->mb, MBE_AlbumGetArtistId, data, MB_BUFFER_SIZE, j)) {
-        mb_GetIDFromURL (priv->mb, data, data, MB_BUFFER_SIZE);
+      if (mb_GetResultData1(priv->mb, MBE_AlbumGetArtistId, data, sizeof (data), j)) {
+        mb_GetIDFromURL (priv->mb, data, data, sizeof (data));
         track->artist_id = g_strdup (data);
       }
 
-      if (mb_GetResultData1(priv->mb, MBE_AlbumGetTrackName, data, MB_BUFFER_SIZE, j)) {
-        track->title = g_strdup (data);
+      if (mb_GetResultData1(priv->mb, MBE_AlbumGetTrackName, data, sizeof (data), j)) {
+        if (track->artist_id != NULL) {
+          artist_and_title_from_title (track, data);
+        } else {
+          track->title = g_strdup (data);
+        }
       }
 
-      if (mb_GetResultData1(priv->mb, MBE_AlbumGetArtistName, data, MB_BUFFER_SIZE, j)) {
+      if (track->artist == NULL && mb_GetResultData1(priv->mb, MBE_AlbumGetArtistName, data, sizeof (data), j)) {
         track->artist = g_strdup (data);
       }
 
-      if (mb_GetResultData1(priv->mb, MBE_AlbumGetTrackDuration, data, MB_BUFFER_SIZE, j)) {
+      if (mb_GetResultData1(priv->mb, MBE_AlbumGetArtistSortName, data, sizeof (data), j)) {
+        track->artist_sortname = g_strdup (data);
+      }
+
+      if (mb_GetResultData1(priv->mb, MBE_AlbumGetTrackDuration, data, sizeof (data), j)) {
         track->duration = atoi (data) / 1000;
       }
       
@@ -432,12 +611,14 @@ lookup_cd (SjMetadata *metadata)
 
     albums = g_list_append (albums, album);
 
-    mb_Select (priv->mb, MBS_Back);
+    mb_Select (priv->mb, MBS_Rewind);
   }
 
   /* For each album, we need to insert the duration data if necessary
    * We need to query this here because otherwise we would flush the
    * data queried from the server */
+  /* TODO: scan for 0 duration before doing the query to avoid another lookup if
+     we don't need to do it */
   mb_Query (priv->mb, MBQ_GetCDTOC);
   for (al = albums; al; al = al->next) {
     AlbumDetails *album = al->data;
@@ -454,6 +635,7 @@ lookup_cd (SjMetadata *metadata)
       j++;
     }
   }
+
   priv->albums = albums;
   g_idle_add ((GSourceFunc)fire_signal_idle, metadata);
   return albums;
@@ -474,3 +656,20 @@ mb_list_albums (SjMetadata *metadata, GError **error)
     return;
   }
 }
+
+static char *
+mb_get_submit_url (SjMetadata *metadata)
+{
+  SjMetadataMusicbrainzPrivate *priv;
+  char url[1025];
+
+  g_return_val_if_fail (metadata != NULL, NULL);
+
+  priv = SJ_METADATA_MUSICBRAINZ (metadata)->priv;
+
+  if (mb_GetWebSubmitURL(priv->mb, url, 1024)) {
+    return g_strdup(url);
+  } else {
+    return NULL;
+  }
+}
